# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# Joanna J. Zou

from time import gmtime, strftime

def log_msg(msg):

    formatted_msg = '{} {}'.format(strftime('%Y-%m-%dT%H:%M:%SZ', gmtime()), msg)

    print(formatted_msg)

log_msg('First line of DL_calculation')

import sys, os, json, ntpath, posixpath, argparse
import numpy as np
import pandas as pd
import shutil

sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))

from pelicun.base import set_options, convert_to_MultiIndex, convert_to_SimpleIndex, describe
from pelicun.assessment import Assessment
from pelicun.file_io import save_to_csv

def add_units(raw_demands, config):

    demands = raw_demands.T

    if "Units" not in demands.index:
        demands.insert(0, "Units", np.nan)
    else:
        return raw_demands

    try:
        length_unit = config['GeneralInformation']['units']['length']
    except:
        log_msg("No units assigned to the raw demand input and default unit "
                "definition missing from the input file. Raw demands cannot "
                "be parsed. Terminating analysis. ")

        return None

    if length_unit == 'in':
        length_unit = 'inch'

    demands = convert_to_MultiIndex(demands, axis=0).sort_index(axis=0).T

    demands.drop(demands.columns[demands.columns.get_level_values(1)== ''], axis=1, inplace=True)

    demands[('1','PRD','1','1')] = demands[('1','PRD','1','1')]
    demands[('1','PFA','3','1')] = demands[('1','PFA','3','1')]

    for EDP_type in ['PFA', 'PGA', 'SA']:
        demands.iloc[0, demands.columns.get_level_values(1) == EDP_type] = length_unit+'ps2'

    for EDP_type in ['PFV', 'PWS', 'PGV', 'SV']:
        demands.iloc[0, demands.columns.get_level_values(1) == EDP_type] = length_unit+'ps'

    for EDP_type in ['PFD', 'PIH', 'SD', 'PGD']:
        demands.iloc[0, demands.columns.get_level_values(1) == EDP_type] = length_unit

    for EDP_type in ['PID', 'PRD', 'DWD', 'RDR', 'PMD', 'RID']:
        demands.iloc[0, demands.columns.get_level_values(1) == EDP_type] = 'rad'

    return convert_to_SimpleIndex(demands, axis=1)

def run_pelicun(config_path, demand_path=None, sample_size=None,
    #DL_method, BIM_file, EDP_file, DM_file, DV_file,
    #output_path=None, detailed_results=True, coupled_EDP=False,
    #log_file=True, event_time=None, ground_failure=False,
    #auto_script_path=None, resource_dir=None
    ):

    config_path = os.path.abspath(config_path)

    with open(config_path, 'r') as f:
        config = json.load(f)

    if demand_path is not None:
        demand_path = os.path.abspath(demand_path) # response.csv

        # check if the raw demands have units identified
        raw_demands = pd.read_csv(demand_path, index_col=0)

        demands = add_units(raw_demands, config)

        if demands is None:
            return -1

        demand_path = demand_path[:-4]+'_ext.csv'

        demands.to_csv(demand_path)

    # Initial setup -----------------------------------------------------------
    general_info = config.get('GeneralInformation', None)
    if general_info is None:
        log_msg("General Information is missing from the input file. "
                "Terminating analysis.")

        return -1

    DL_config = config.get('DamageAndLoss', None)
    if DL_config is None:

        log_msg("Damage and Loss configuration missing from input file. "
                "Terminating analysis.")

        return -1

    PAL = Assessment(DL_config.get("Options", None))

    if general_info.get('NumberOfStories', False):
        PAL.stories = general_info['NumberOfStories']

    if sample_size is None:
        if DL_config.get("SampleSize", False):
            sample_size = DL_config["SampleSize"]

    asset_config = DL_config.get('Asset', None)
    demand_config = DL_config.get('Demand', None)
    damage_config = DL_config.get('Damage', None)
    loss_config = DL_config.get('Loss', None)
    out_config = DL_config.get('Outputs', None)

    if asset_config is None:
        log_msg("Asset configuration missing. Terminating analysis.")
        return -1

    if demand_config is None:
        log_msg("Demand configuration missing. Terminating analysis.")
        return -1

    if out_config is None:
        log_msg("Output configuration missing. Terminating analysis.")
        return -1

    # Asset Definition ------------------------------------------------------------

    # if requested, load a component model and generate a sample
    if asset_config.get('LoadModelFrom', False):

        if sample_size is None:
            log_msg("Sample size not specified. Terminating analysis.")
            return -1

        # load component model
        PAL.asset.load_cmp_model(
            data_source= asset_config['LoadModelFrom'])

        # generate component quantity sample
        PAL.asset.generate_cmp_sample(sample_size)

    # if requested, load the quantity sample from a file
    if asset_config.get('LoadSampleFrom', False):
        PAL.asset.load_cmp_sample(asset_config['LoadSampleFrom'])

    # if requested, save results
    if out_config.get('Asset', False):

        out_reqs = [out if val else "" for out, val in out_config['Asset'].items()]

        if np.any(np.isin(['Sample', 'Statistics'], out_reqs)):
            cmp_sample = PAL.asset.save_cmp_sample()

            if 'Sample' in out_reqs:
                cmp_sample_s = convert_to_SimpleIndex(cmp_sample, axis=1)
                cmp_sample_s.to_csv("CMP_sample.zip",
                                    index_label=cmp_sample_s.columns.name,
                                    compression=dict(method='zip',
                                                     archive_name='CMP_sample.csv'))

            if 'Statistics' in out_reqs:
                cmp_stats = convert_to_SimpleIndex(describe(cmp_sample), axis=1)
                cmp_stats.to_csv("CMP_stats.csv",
                                    index_label=cmp_stats.columns.name)

    # Demand Assessment -----------------------------------------------------------

    # if demand calibration is requested
    if demand_config.get('Calibration', False):

        cal_config = demand_config['Calibration']

        # load demand samples to serve as reference data
        if cal_config.get('LoadSampleFrom', False):
            demand_path = cal_config['LoadSampleFrom']

        if demand_path is None:
            log_msg("Demand sample not specified for demand calibration."
                    "Terminating analysis.")
            return -1

        PAL.demand.load_sample(demand_path)

        # then use it to calibrate the demand model
        PAL.demand.calibrate_model(cal_config['Marginals'])

        # if requested, save the model to files
        if cal_config.get('SaveModelTo', False):
            PAL.demand.save_model(file_prefix=cal_config['SaveModelTo'])

    # if requested, load a pre-calibrated model from files
    if demand_config.get('LoadModelFrom', False):
        PAL.demand.load_model(data_source= demand_config['LoadModelFrom'])

    # if demand samples are provided in a file
    if demand_config.get('LoadSampleFrom', False):
        PAL.demand.load_sample(demand_config['LoadSampleFrom'])

    else:
        # otherwise, generate demand sample
        if sample_size is None:
            log_msg("Sample size not specified. Terminating analysis.")
            return -1

        PAL.demand.generate_sample({"SampleSize": sample_size})

    # if requested, save results
    if out_config.get('Demand', False):

        out_reqs = [out if val else "" for out, val in out_config['Demand'].items()]

        if np.any(np.isin(['Sample', 'Statistics'], out_reqs)):
            demand_sample = PAL.demand.save_sample()

            if 'Sample' in out_reqs:
                demand_sample_s = convert_to_SimpleIndex(demand_sample, axis=1)
                demand_sample_s.to_csv("DEM_sample.zip",
                                       index_label=demand_sample_s.columns.name,
                                       compression=dict(method='zip',
                                                        archive_name='DEM_sample.csv'))

            if 'Statistics' in out_reqs:
                demand_stats = convert_to_SimpleIndex(describe(demand_sample), axis=1)
                demand_stats.to_csv("DEM_stats.csv",
                                    index_label=demand_stats.columns.name)

    # Damage Assessment -----------------------------------------------------------

    # if a damage assessment is requested
    if damage_config is not None:

        if sample_size is None:
            log_msg("Sample size not specified. Terminating analysis.")
            return -1

        # load the fragility information
        PAL.damage.load_damage_model(damage_config['LoadModelFrom'])

        # calculate damages
        # load the damage process if needed
        if damage_config.get('DamageProcess', False):
            with open(damage_config['DamageProcess'], 'r') as f:
                dmg_process = json.load(f)
        else:
            dmg_process = None

        PAL.damage.calculate(sample_size, dmg_process=dmg_process)

        # if damage samples are provided in a file
        if damage_config.get('LoadSampleFrom', False):
            PAL.damage.load_sample(damage_config['LoadSampleFrom'])

        # if requested, save results
        if out_config.get('Damage', False):

            out_reqs = [out if val else ""
                        for out, val in out_config['Damage'].items()]

            if np.any(np.isin(['Sample', 'Statistics',
                               'GroupedSample', 'GroupedStatistics'],
                              out_reqs)):
                damage_sample = PAL.damage.save_sample()

                if 'Sample' in out_reqs:
                    damage_sample_s = convert_to_SimpleIndex(damage_sample, axis=1)
                    damage_sample_s.to_csv(
                        "DMG_sample.zip",
                        index_label=damage_sample_s.columns.name,
                        compression=dict(method='zip',
                                         archive_name='DMG_sample.csv'))

                if 'Statistics' in out_reqs:
                    damage_stats = convert_to_SimpleIndex(describe(damage_sample),
                                                          axis=1)
                    damage_stats.to_csv("DMG_stats.csv",
                                        index_label=damage_stats.columns.name)

                if np.any(np.isin(['GroupedSample', 'GroupedStatistics'], out_reqs)):
                    grp_damage = damage_sample.groupby(level=[0,3], axis=1).sum()

                    if 'GroupedSample' in out_reqs:
                        grp_damage_s = convert_to_SimpleIndex(grp_damage, axis=1)
                        grp_damage_s.to_csv("DMG_grp.zip",
                                            index_label=grp_damage_s.columns.name,
                                            compression=dict(
                                                method='zip',
                                                archive_name='DMG_grp.csv'))

                    if 'GroupedStatistics' in out_reqs:
                        grp_stats = convert_to_SimpleIndex(describe(grp_damage),
                                                           axis=1)
                        grp_stats.to_csv("DMG_grp_stats.csv",
                                         index_label=grp_stats.columns.name)

    # Loss Assessment -----------------------------------------------------------

    # if a loss assessment is requested
    if loss_config is not None:

        out_config_loss = out_config.get('Loss', {})

        loss_map_path = loss_config['LoadMappingFrom']

        # if requested, calculate repair consequences
        if loss_config.get('CalculateBldgRepair', False):

            bldg_repair_config = loss_config['CalculateBldgRepair']

            PAL.bldg_repair.load_model(bldg_repair_config['LoadModelFrom'],
                                       loss_map_path)

            PAL.bldg_repair.calculate(sample_size)

            agg_repair = PAL.bldg_repair.aggregate_losses()

            if out_config_loss.get('BldgRepair', False):

                out_reqs = [out if val else ""
                            for out, val in out_config_loss['BldgRepair'].items()]

                if np.any(np.isin(['Sample', 'Statistics',
                                   'GroupedSample', 'GroupedStatistics',
                                   'AggregateSample', 'AggregateStatistics'],
                                  out_reqs)):
                    repair_sample = PAL.bldg_repair.save_sample()

                    if 'Sample' in out_reqs:
                        repair_sample_s = convert_to_SimpleIndex(repair_sample, axis=1)
                        repair_sample_s.to_csv(
                            "DV_bldg_repair_sample.zip",
                            index_label=repair_sample_s.columns.name,
                            compression=dict(method='zip',
                                             archive_name='DV_bldg_repair_sample.csv'))

                    if 'Statistics' in out_reqs:
                        repair_stats = convert_to_SimpleIndex(describe(repair_sample),
                                                              axis=1)
                        repair_stats.to_csv("DV_bldg_repair_stats.csv",
                                            index_label=repair_stats.columns.name)

                    if np.any(np.isin(['GroupedSample', 'GroupedStatistics'], out_reqs)):
                        grp_repair = repair_sample.groupby(level=[0,1,2], axis=1).sum()

                        if 'GroupedSample' in out_reqs:
                            grp_repair_s = convert_to_SimpleIndex(grp_repair, axis=1)
                            grp_repair_s.to_csv("DV_bldg_repair_grp.zip",
                                              index_label=grp_repair_s.columns.name,
                                              compression=dict(
                                                method='zip',
                                                archive_name='DV_bldg_repair_grp.csv'))

                        if 'GroupedStatistics' in out_reqs:
                            grp_stats = convert_to_SimpleIndex(describe(grp_repair), axis=1)
                            grp_stats.to_csv("DV_bldg_repair_grp_stats.csv",
                                             index_label=grp_stats.columns.name)

                    if np.any(np.isin(['AggregateSample', 'AggregateStatistics'], out_reqs)):

                        if 'AggregateSample' in out_reqs:
                            agg_repair_s = convert_to_SimpleIndex(agg_repair, axis=1)
                            agg_repair_s.to_csv("DV_bldg_repair_agg.zip",
                                              index_label=agg_repair_s.columns.name,
                                              compression=dict(
                                                method='zip',
                                                archive_name='DV_bldg_repair_agg.csv'))

                        if 'AggregateStatistics' in out_reqs:
                            agg_stats = convert_to_SimpleIndex(describe(agg_repair), axis=1)
                            agg_stats.to_csv("DV_bldg_repair_agg_stats.csv",
                                             index_label=agg_stats.columns.name)

    return 0

def main(args):

    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--configFile')
    parser.add_argument('-d', '--demandFile', default = None)
    parser.add_argument('-s', '--sampleSize', default = None)
    #parser.add_argument('--DL_Method', default = None)
    #parser.add_argument('--outputBIM', default='BIM.csv')
    #parser.add_argument('--outputEDP', default='EDP.csv')
    #parser.add_argument('--outputDM', default = 'DM.csv')
    #parser.add_argument('--outputDV', default = 'DV.csv')
    #parser.add_argument('--dirnameOutput', default = None)
    #parser.add_argument('--event_time', default=None)
    #parser.add_argument('--detailed_results', default = True,
    #    type = str2bool, nargs='?', const=True)
    #parser.add_argument('--coupled_EDP', default = False,
    #    type = str2bool, nargs='?', const=False)
    #parser.add_argument('--log_file', default = True,
    #    type = str2bool, nargs='?', const=True)
    #parser.add_argument('--ground_failure', default = False,
    #    type = str2bool, nargs='?', const=False)
    #parser.add_argument('--auto_script', default=None)
    #parser.add_argument('--resource_dir', default=None)
    args = parser.parse_args(args)

    log_msg('Initializing pelicun calculation...')

    #print(args)
    out = run_pelicun(
        args.configFile, args.demandFile,
        args.sampleSize,
        #args.DL_Method,
        #args.outputBIM, args.outputEDP,
        #args.outputDM, args.outputDV,
        #output_path = args.dirnameOutput,
        #detailed_results = args.detailed_results,
        #coupled_EDP = args.coupled_EDP,
        #log_file = args.log_file,
        #event_time = args.event_time,
        #ground_failure = args.ground_failure,
        #auto_script_path = args.auto_script,
        #resource_dir = args.resource_dir
    )

    if out == -1:
        log_msg("pelicun calculation failed.")
    else:
        log_msg('pelicun calculation completed.')

if __name__ == '__main__':

    main(sys.argv[1:])



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay

"""
This module defines constants, basic classes and methods for pelicun.

.. rubric:: Contents

.. autosummary::

    set_options
    convert_to_SimpleIndex
    convert_to_MultiIndex
    convert_unit
    show_matrix
    show_warning
    print_system_info
    log_div
    log_msg
    describe
    str2bool

    Options

"""

import os, sys, time
import warnings
from datetime import datetime
from time import strftime
from pathlib import Path
import argparse

from copy import deepcopy

# import libraries for other modules
import numpy as np
import pandas as pd

from .__init__ import __version__ as pelicun_version

idx = pd.IndexSlice

# set printing options
import pprint
pp = pprint.PrettyPrinter(indent=2, width=80-24)

pd.options.display.max_rows = 20
pd.options.display.max_columns = None
pd.options.display.expand_frame_repr = True
pd.options.display.width = 300

idx = pd.IndexSlice

class Options(object):

    """

    Parameters

    verbose: boolean
        If True, the pelicun echoes more information throughout the assessment.
        This can be useful for debugging purposes.

    log_show_ms: boolean
        If True, the timestamps in the log file are in microsecond precision.

    """

    def __init__(self):

        self._verbose = False
        self._log_show_ms = False
        self._print_log = False

        self.defaults = None
        self.sampling_method = None

        self._seed = None
        self._rng = np.random.default_rng()

        self.reset_log_strings()

        self.demand_offset = {}
        self.nondir_multi_dict = {}

    def nondir_multi(self, EDP_type):

        if EDP_type in self.nondir_multi_dict.keys():
            return self.nondir_multi_dict[EDP_type]

        elif 'ALL' in self.nondir_multi_dict.keys():
            return self.nondir_multi_dict['ALL']

        else:
            raise ValueError(f"Scale factor for non-directional demand "
                             f"calculation of {EDP_type} not specified.")

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = bool(value)

    @property
    def log_show_ms(self):
        return self._log_show_ms

    @log_show_ms.setter
    def log_show_ms(self, value):
        self._log_show_ms = bool(value)

        self.reset_log_strings()

    @property
    def log_pref(self):
        return self._log_pref

    @property
    def log_div(self):
        return self._log_div

    @property
    def log_time_format(self):
        return self._log_time_format

    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, value):
        self._seed = value

        self._rng = np.random.default_rng(self._seed)

    @property
    def rng(self):
        return self._rng

    @property
    def log_file(self):
        return globals()['log_file']

    @log_file.setter
    def log_file(self, value):

        if value is None:
            globals()['log_file'] = value

        else:

            try:

                filepath = Path(value).resolve()

                globals()['log_file'] = str(filepath)

                with open(filepath, 'w') as f:
                    f.write('')

            except:
                log_msg(f"WARNING: The filepath provided for the log file does "
                        f"not point to a valid location: {value}. \nPelicun "
                        f"cannot print the log to a file.")
                self.log_file = None

    @property
    def print_log(self):
        return self._print_log

    @print_log.setter
    def print_log(self, value):
        self._print_log = str2bool(value)

    def reset_log_strings(self):

        if self._log_show_ms:
            self._log_time_format = '%H:%M:%S:%f'
            self._log_pref = ' ' * 16 # the length of the time string in the log file
            self._log_div = '-' * (80 - 17) # to have a total length of 80 with the time added
        else:
            self._log_time_format = '%H:%M:%S'
            self._log_pref = ' ' * 9
            self._log_div = '-' * (80 - 10)

    def scale_factor(self, unit):

        if unit is not None:

            if unit in globals().keys():
                scale_factor = globals()[unit]

            else:
                raise ValueError(f"Unknown unit: {unit}")
        else:
            scale_factor = 1.0

        return scale_factor

options = Options()

log_file = None

# get the absolute path of the pelicun directory
pelicun_path = Path(os.path.dirname(os.path.abspath(__file__)))

def set_options(config_options):

    if config_options is not None:

        for key, value in config_options.items():

            if key == "Verbose":
                options.verbose = value
            elif key == "Seed":
                options.seed = value
            elif key == "LogShowMS":
                options.log_show_ms = value
            elif key == "LogFile":
                options.log_file = value
            elif key == "PrintLog":
                options.print_log = value
            elif key == "SamplingMethod":
                options.sampling_method = value
            elif key == "DemandOffset":
                options.demand_offset = value
            elif key == "NonDirectionalMultipliers":
                options.nondir_multi_dict = value
            elif key == "RepairCostAndTimeCorrelation":
                options.rho_cost_time = value
            elif key == "EconomiesOfScale":
                options.eco_scale = value

def convert_to_SimpleIndex(data, axis=0, inplace=False):
    """
    Converts the index of a DataFrame to a simple, one-level index

    The target index uses standard SimCenter convention to identify different
    levels: a dash character ('-') is used to separate each level of the index.

    Parameters
    ----------
    data: DataFrame
        The DataFrame that will be modified.
    axis: int, optional, default:0
        Identifies if the index (0) or the columns (1) shall be edited.
    inplace: bool, optional, default:False
        If yes, the operation is performed directly on the input DataFrame
        and not on a copy of it.

    Returns
    -------
    data: DataFrame
        The modified DataFrame
    """

    if axis in [0, 1]:

        if inplace:
            data_mod = data
        else:
            data_mod = data.copy()

        if axis == 0:
            simple_name = '-'.join([n if n is not None else "" for n in data.index.names])
            simple_index = ['-'.join([str(id_i) for id_i in id])
                            for id in data.index]

            data_mod.index = simple_index
            data_mod.index.name = simple_name

        elif axis == 1:
            simple_name = '-'.join([n if n is not None else "" for n in data.columns.names])
            simple_index = ['-'.join([str(id_i) for id_i in id])
                            for id in data.columns]

            data_mod.columns = simple_index
            data_mod.columns.name = simple_name

    else:
        raise ValueError(f"Invalid axis parameter: {axis}")

    return data_mod

def convert_to_MultiIndex(data, axis=0, inplace=False):
    """
    Converts the index of a DataFrame to a MultiIndex

    We assume that the index uses standard SimCenter convention to identify
    different levels: a dash character ('-') is expected to separate each level
    of the index.

    Parameters
    ----------
    data: DataFrame
        The DataFrame that will be modified.
    axis: int, optional, default:0
        Identifies if the index (0) or the columns (1) shall be edited.
    inplace: bool, optional, default:False
        If yes, the operation is performed directly on the input DataFrame
        and not on a copy of it.

    Returns
    -------
    data: DataFrame
        The modified DataFrame
    """

    # check if the requested axis is already a MultiIndex
    if (((axis == 0) and (isinstance(data.index, pd.MultiIndex))) or
        ((axis == 1) and (isinstance(data.columns, pd.MultiIndex)))):

        # if yes, return the data unchanged
        return data

    if axis == 0:
        index_labels = [str(label).split('-') for label in data.index]

    elif axis == 1:
        index_labels = [str(label).split('-') for label in data.columns]

    else:
        raise ValueError(f"Invalid axis parameter: {axis}")

    max_lbl_len = np.max([len(labels) for labels in index_labels])

    for l_i, labels in enumerate(index_labels):

        if len(labels) != max_lbl_len:
            labels += ['', ] * (max_lbl_len - len(labels))
            index_labels[l_i] = labels

    index_labels = np.array(index_labels)

    if index_labels.shape[1] > 1:

        if inplace:
            data_mod = data
        else:
            data_mod = data.copy()

        if axis == 0:
            data_mod.index = pd.MultiIndex.from_arrays(index_labels.T)

        else:
            data_mod.columns = pd.MultiIndex.from_arrays(index_labels.T)

        return data_mod

    else:
        return data

def calc_unit_scale_factor(unit):
    """
    Determines the scale factor from input unit to the corresponding SI unit

    Parameters
    ----------
    unit: str
        Either a unit name, or a quantity and a unit name separated by a space.
        For example: 'ft' or '100 ft'.

    Returns
    -------
    scale_factor: float
        Scale factor that convert values from unit to SI unit
    """

    unit = unit.strip().split(' ')

    # check if there is a quantity specified; if yes, parse it
    if len(unit) > 1:
        unit_count, unit_name = unit
        unit_count = float(unit_count)

    else:
        unit_count = 1
        unit_name = unit[0]

    try:
        scale_factor = unit_count * globals()[unit_name]

    except:
        raise ValueError(f"Specified unit not recognized: "
                         f"{unit_count} {unit_name}")

    return scale_factor

# print a matrix in a nice way using a DataFrame
def show_matrix(data, describe=False):
    if describe:
        pp.pprint(pd.DataFrame(data).describe(percentiles=[0.01,0.1,0.5,0.9,0.99]))
    else:
        pp.pprint(pd.DataFrame(data))

# Monkeypatch warnings to get prettier messages
def _warning(message, category, filename, lineno, file=None, line=None):

    if '\\' in filename:
        file_path = filename.split('\\')
    elif '/' in filename:
        file_path = filename.split('/')
    else:
        file_path = None

    if file_path is not None:
        python_file = '/'.join(file_path[-3:])
    else:
        python_file = filename

    print('WARNING in {} at line {}\n{}\n'.format(python_file, lineno, message))

warnings.showwarning = _warning

def show_warning(warning_msg):
    warnings.warn(UserWarning(warning_msg))

def print_system_info():

    log_msg('System Information:',
            prepend_timestamp=False, prepend_blank_space=False)
    log_msg(f'local time zone: {datetime.utcnow().astimezone().tzinfo}\n'
            f'start time: {datetime.now().strftime("%Y-%m-%dT%H:%M:%S")}\n'
            f'python: {sys.version}\n'
            f'numpy: {np.__version__}\n'
            f'pandas: {pd.__version__}\n',
            prepend_timestamp=False)

def log_div(prepend_timestamp=False):
    """
    Print a divider line to the log file

    """

    if prepend_timestamp:
        msg = options.log_div

    else:
        msg = '-' * 80

    log_msg(msg, prepend_timestamp = prepend_timestamp)


def log_msg(msg='', prepend_timestamp=True, prepend_blank_space=True):
    """
    Print a message to the screen with the current time as prefix

    The time is in ISO-8601 format, e.g. 2018-06-16T20:24:04Z

    Parameters
    ----------
    msg: string
       Message to print.

    """

    msg_lines = msg.split('\n')

    for msg_i, msg_line in enumerate(msg_lines):

        if (prepend_timestamp and (msg_i==0)):
            formatted_msg = '{} {}'.format(
                datetime.now().strftime(options.log_time_format), msg_line)
        elif prepend_timestamp:
            formatted_msg = options.log_pref + msg_line
        elif prepend_blank_space:
            formatted_msg = options.log_pref + msg_line
        else:
            formatted_msg = msg_line

        if options.print_log:
            print(formatted_msg)

        if globals()['log_file'] is not None:
            with open(globals()['log_file'], 'a') as f:
                f.write('\n'+formatted_msg)

def describe(df, percentiles=[0.001, 0.023, 0.10, 0.159, 0.5, 0.841, 0.90,
                              0.977, 0.999]):

    if not isinstance(df, (pd.Series, pd.DataFrame)):
        vals = df
        cols = np.arange(vals.shape[1]) if vals.ndim > 1 else 0

        if vals.ndim == 1:
            df = pd.Series(vals, name=cols)
        else:
            df = pd.DataFrame(vals, columns = cols)

    desc = df.describe(percentiles).T

    # add log standard deviation to the stats
    desc.insert(3, "log_std", np.nan)
    desc = desc.T

    for col in desc.columns:
        if np.min(df[col])>0.0:
            desc.loc['log_std', col] = np.std(np.log(df[col]))

    return desc

def str2bool(v):
    # courtesy of Maxim @ stackoverflow

    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 'True', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'False', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

# Constants for unit conversion

# time
sec = 1.

minute = 60. * sec
hour = 60. * minute
day = 24. * hour

sec2 = sec**2.

# distance, area, volume
m = 1.

mm = 0.001 * m
cm = 0.01 * m
km = 1000. * m

inch = 0.0254
ft = 12. * inch
mile = 5280. * ft

# area
m2 = m**2.

mm2 = mm**2.
cm2 = cm**2.
km2 = km**2.

inch2 = inch**2.
ft2 = ft**2.
mile2 = mile**2.

# volume
m3 = m**3.

inch3 = inch**3.
ft3 = ft**3.


# speed / velocity
cmps = cm / sec
mps = m / sec
mph = mile / hour

inchps = inch / sec
ftps = ft / sec

# acceleration
mps2 = m / sec2

inchps2 = inch / sec2
ftps2 = ft / sec2

g = 9.80665 * mps2

# mass
kg = 1.

ton = 1000. * kg

lb = 0.453592 * kg

# force
N = kg * m / sec2

kN = 1e3 * N

lbf = lb * g
kip = 1000. * lbf
kips = kip

# pressure / stress
Pa = N / m2

kPa = 1e3 * Pa
MPa = 1e6 * Pa
GPa = 1e9 * Pa

psi = lbf / inch2
ksi = 1e3 * psi
Mpsi = 1e6 * psi

# misc
A = 1.

V = 1.
kV = 1000. * V

ea = 1.

rad = 1.

C = 1.

USD_2011 = 1.
worker_day = 1.

# FEMA P58 specific
#TODO: work around these and make them available only in the parser methods
EA = ea
SF = ft2
LF = ft
TN = ton
AP = A
CF = ft3 / minute
KV = kV * A

# Input specs

CMP_data_path = dict(
    P58      = '/resources/FEMA_P58_2nd_ed.hdf',
    HAZUS_EQ = '/resources/HAZUS_MH_2.1_EQ.hdf',
    HAZUS_HU = '/resources/HAZUS_MH_2.1.hdf',
    HAZUS_FL = '/resources/HAZUS_MH_2.1_FL.hdf',
    HAZUS_MISC = '/resources/HAZUS_MH_2.1_MISC.hdf'
)

POP_data_path = dict(
    P58      = '/resources/FEMA_P58_2nd_ed.hdf',
    HAZUS_EQ = '/resources/HAZUS_MH_2.1_EQ.hdf'
)

default_units = dict(
    force =        'N',
    length =       'm',
    area =         'm2',
    volume =       'm3',
    speed =        'mps',
    acceleration = 'mps2',
)

EDP_units = dict(
    # drifts and rotations are not listed here because they are unitless

    # Floor response
    PFA = 'acceleration',
    PFV = 'speed',
    PFD = 'length',

    # Wind intensity
    PWS = 'speed',

    # Inundation intensity
    PIH = 'length',

    # Shaking intensity
    PGA = 'acceleration',
    PGV = 'speed',
    SA = 'acceleration',
    SV = 'speed',
    SD = 'length',
    PGD = 'length',
)

EDP_to_demand_type = {
    # Drifts
    'Story Drift Ratio' :             'PID',
    'Peak Interstory Drift Ratio':    'PID',
    'Roof Drift Ratio' :              'PRD',
    'Peak Roof Drift Ratio' :         'PRD',
    'Damageable Wall Drift' :         'DWD',
    'Racking Drift Ratio' :           'RDR',
    'Mega Drift Ratio' :              'PMD',
    'Residual Drift Ratio' :          'RID',
    'Residual Interstory Drift Ratio':'RID',
    'Peak Effective Drift Ratio':     'EDR',

    # Floor response
    'Peak Floor Acceleration' :       'PFA',
    'Peak Floor Velocity' :           'PFV',
    'Peak Floor Displacement':        'PFD',

    # Component response
    'Peak Link Rotation Angle':       'LR',
    'Peak Link Beam Chord Rotation':  'LBR',

    # Wind Intensity
    'Peak Gust Wind Speed' :          'PWS',

    # Inundation Intensity
    'Peak Inundation Height' :        'PIH',

    # Shaking Intensity
    'Peak Ground Acceleration' :      'PGA',
    'Peak Ground Velocity' :          'PGV',
    'Spectral Acceleration' :         'SA',
    'Spectral Velocity' :             'SV',
    'Spectral Displacement' :         'SD',
    'Peak Spectral Acceleration' :    'SA',
    'Peak Spectral Velocity' :        'SV',
    'Peak Spectral Displacement' :    'SD',
    'Permanent Ground Deformation' :  'PGD',
}

# PFA in FEMA P58 corresponds to the top of the given story. The ground floor
# has an index of 0. When damage of acceleration-sensitive components
# is controlled by the acceleration of the bottom of the story, the
# corresponding PFA location needs to be reduced by 1. The SimCenter framework
# assumes that PFA corresponds to the bottom of the given story
# by default, hence, we would need to subtract 1 from the location values.
# Rather than changing the locations themselves, we assign an offset of -1
# so that the results still get collected at the appropriate story.
EDP_offset_adjustment = dict(
    PFA = -1,
    PFV = -1,
    PFD = -1
)


# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of the SimCenter Backend Applications
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# this file. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# Kuanshi Zhong

"""
These are unit and integration tests on the auto_HU_NJ module.
"""

import pytest
import numpy as np
from numpy.testing import assert_allclose
import json
import os
import inspect
# Importing auto_HU_NJ module
from auto_HU_NJ import *
# Current directory
cur_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# Test input directory
base_input_path = 'resources'

def test_parse_BIM():
    """
    Testing the parse_BIM function.
    """
    # Testing the ruleset for Hurricane-Prone Region (HPR)
    res = []
    ref = [1, 0, 1, 0]
    for i in range(4):
        BIM_dir = os.path.join(cur_dir, base_input_path, 'BIM_Data',
                               'parse_BIM_test_' + str(i+1) + '.json')
        with open(BIM_dir) as f:
            BIM_input = json.load(f)
        BIM_output = parse_BIM(BIM_input['GI'])
        res.append(int(BIM_output['HPR']))
    # Check
    assert_allclose(res, ref, atol=1e-5)

    # Testing the ruleset for Wind Borne Debris (WBD)
    res = []
    ref = [0, 0, 0, 0, 1, 1, 1, 1]
    for i in range(8):
        BIM_dir = os.path.join(cur_dir, base_input_path, 'BIM_Data',
                               'parse_BIM_test_' + str(i+1) + '.json')
        with open(BIM_dir) as f:
            BIM_input = json.load(f)
        BIM_output = parse_BIM(BIM_input['GI'])
        res.append(int(BIM_output['WBD']))
    # Check
    assert_allclose(res, ref, atol=1e-5)

    # Testing the ruleset for terrain
    res = []
    ref = [3, 15, 35, 70, 3, 15, 35, 70]
    for i in range(8):
        BIM_dir = os.path.join(cur_dir, base_input_path, 'BIM_Data',
                               'parse_BIM_test_' + str(i+1) + '.json')
        with open(BIM_dir) as f:
            BIM_input = json.load(f)
        BIM_output = parse_BIM(BIM_input['GI'])
        res.append(int(BIM_output['terrain']))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_building_class():
    """
    Testing the building class function.
    """
    # Testing the ruleset for classifying Hazus building class
    res = []
    ref_class = ['WSF', 'WMUH', 'SERB', 'SECB', 'SPMB', 'CERB', 'CECB', 'MSF',
                 'MERB', 'MECB', 'MLRI', 'MMUH', 'MLRM']
    ref = np.ones(13)
    for i in range(13):
        data_dir = os.path.join(cur_dir, base_input_path, 'BuildingClass_Data',
                               'building_class_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = building_class(tmp)
        print(data_output)
        res.append(int(data_output == ref_class[i]))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_WSF_config():
    """
    Testing the WSF_config function.
    """
    res = []
    ref_class = ['WSF2_gab_0_8d_tnail_no',
                 'WSF2_gab_1_8d_tnail_no',
                 'WSF2_hip_1_8d_tnail_no',
                 'WSF2_hip_0_8d_tnail_no',
                 '8s_strap_no',
                 '8s_strap_no',
                 '8s_tnail_no',
                 '8s_strap_sup',
                 '8d_strap_std',
                 '8d_tnail_wkd',
                 'WSF1']
    ref = np.ones(11)
    for i in range(11):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'wsf_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = WSF_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_WMUH_config():
    """
    Testing the WMUH_config function.
    """
    res = []
    ref_class = ['WMUH2_flt_spm_god_null',
                 'WMUH2_flt_spm_god_null',
                 'WMUH2_flt_spm_god_null',
                 'WMUH2_gab_null_null_1',
                 'WMUH2_hip_null_null_1',
                 'WMUH2_gab_null_null_0',
                 'WMUH2_hip_null_null_0',
                 'WMUH2_flt_spm_por_null',
                 'WMUH2_flt_bur_por_null',
                 'WMUH2_flt_spm_god_null_8s',
                 'WMUH2_flt_spm_god_null_8d',
                 'WMUH2_flt_spm_god_null_8s',
                 'WMUH2_flt_spm_god_null_8d',
                 'strap',
                 'tnail',
                 'tnail',
                 'tnail_1',
                 'WMUH3']
    ref = np.ones(18)
    for i in range(18):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'wmuh_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = WMUH_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_MSF_config():
    """
    Testing the MSF_config function.
    """
    res = []
    ref_class = ['nav_1',
                 'nav_0',
                 '8s',
                 '8d',
                 '8s',
                 '8d',
                 'MSF2']
    ref = np.ones(7)
    for i in range(7):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'msf_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = MSF_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_MMUH_config():
    """
    Testing the MMUH_config function.
    """
    res = []
    ref_class = ['flt_1_spm_god',
                 'flt_1_spm_por',
                 'flt_1_bur_por',
                 '8s_strap',
                 '8d_strap',
                 '8d_tnail',
                 '8s_strap',
                 '8d_tnail',
                 'MMUH3']
    ref = np.ones(9)
    for i in range(9):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'mmuh_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = MMUH_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_MLRM_config():
    """
    Testing the MLRM_config function.
    """
    res = []
    ref_class = ['spm',
                 'bur',
                 'C',
                 'D',
                 'A',
                 '6d_god',
                 '6d_por',
                 'std',
                 'sup',
                 'A_1_sup',
                 'sgl',
                 'mlt']
    ref = np.ones(12)
    for i in range(12):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'mlrm_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = MLRM_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_MLRI_config():
    """
    Testing the MLRI_config function.
    """
    res = []
    ref_class = ['sup',
                 'std',
                 'god',
                 'por',
                 'god',
                 'por']
    ref = np.ones(6)
    for i in range(6):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'mlri_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = MLRI_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_MERB_config():
    """
    Testing the MERB_config function.
    """
    res = []
    ref_class = ['bur',
                 'spm',
                 'bur',
                 'C',
                 'D',
                 'A',
                 'std',
                 'sup',
                 'low',
                 'med',
                 'hig',
                 'MERBL',
                 'MERBM',
                 'MERBH']
    ref = np.ones(14)
    for i in range(14):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'merb_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = MERB_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_MECB_config():
    """
    Testing the MECB_config function.
    """
    res = []
    ref_class = ['bur',
                 'spm',
                 'bur',
                 'C',
                 'D',
                 'A',
                 'std',
                 'sup',
                 'low',
                 'med',
                 'hig',
                 'MECBL',
                 'MECBM',
                 'MECBH']
    ref = np.ones(14)
    for i in range(14):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'mecb_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = MECB_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_CECB_config():
    """
    Testing the CECB_config function.
    """
    res = []
    ref_class = ['bur',
                 'spm',
                 'bur',
                 'C',
                 'D',
                 'A',
                 'low',
                 'med',
                 'hig',
                 'CECBL',
                 'CECBM',
                 'CECBH']
    ref = np.ones(12)
    for i in range(12):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'cecb_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = CECB_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_CERB_config():
    """
    Testing the CERB_config function.
    """
    res = []
    ref_class = ['bur',
                 'spm',
                 'bur',
                 'C',
                 'D',
                 'A',
                 'low',
                 'med',
                 'hig',
                 'CERBL',
                 'CERBM',
                 'CERBH']
    ref = np.ones(12)
    for i in range(12):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'cerb_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = CERB_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_SPMB_config():
    """
    Testing the SPMB_config function.
    """
    res = []
    ref_class = ['god',
                 'por',
                 'std',
                 'sup',
                 'SPMBS',
                 'SPMBM',
                 'SPMBL']
    ref = np.ones(7)
    for i in range(7):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'spmb_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = SPMB_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_SECB_config():
    """
    Testing the SECB_config function.
    """
    res = []
    ref_class = ['bur',
                 'spm',
                 'bur',
                 'C',
                 'D',
                 'A',
                 'std',
                 'sup',
                 'low',
                 'med',
                 'hig',
                 'SECBL',
                 'SECBM',
                 'SECBH']
    ref = np.ones(14)
    for i in range(14):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'secb_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = SECB_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_SERB_config():
    """
    Testing the SERB_config function.
    """
    res = []
    ref_class = ['bur',
                 'spm',
                 'bur',
                 'C',
                 'D',
                 'A',
                 'std',
                 'sup',
                 'low',
                 'med',
                 'hig',
                 'SERBL',
                 'SERBM',
                 'SERBH']
    ref = np.ones(14)
    for i in range(14):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'secb_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = SERB_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_FL_config():
    """
    Testing the FL_config function.
    """
    res = []
    ref_class = ['raz',
                 'cvz',
                 'sl',
                 'bn',
                 'bw']
    ref = np.ones(5)
    for i in range(5):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'fl_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        data_output = FL_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)


def test_Assm_config():
    """
    Testing the Assm_config function.
    """
    res = []
    ref_class = ['raz',
                 'caz',
                 'cvz',
                 '1',
                 '0']
    ref = np.ones(5)
    for i in range(5):
        data_dir = os.path.join(cur_dir, base_input_path, 'Config_Data',
                               'assm_test_' + str(i+1) + '.json')
        with open(data_dir) as f:
            data_input = json.load(f)
        tmp = parse_BIM(data_input['GI'])
        tmp2, data_output = Assm_config(tmp)
        print(data_output)
        res.append(int(ref_class[i] in data_output))
    # Check
    assert_allclose(res, ref, atol=1e-5)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay

"""
This module has classes and methods that auto-populate DL models.

.. rubric:: Contents

.. autosummary::

    auto_populate

"""

from .base import *
import importlib
import json
from pathlib import Path

ap_DesignLevel = {
    1940: 'Pre-Code',
    1940: 'Low-Code',
    1975: 'Moderate-Code',
    2100: 'High-Code'
}

ap_DesignLevel_W1 = {
       0: 'Pre-Code',
       0: 'Low-Code',
    1975: 'Moderate-Code',
    2100: 'High-Code'
}

ap_Occupancy = {
    'Other/Unknown': 'RES3',
    'Residential - Single-Family': 'RES1',
    'Residential - Town-Home': 'RES3',
    'Residential - Multi-Family': 'RES3',
    'Residential - Mixed Use': 'RES3',
    'Office': 'COM4',
    'Hotel': 'RES4',
    'School': 'EDU1',
    'Industrial - Light': 'IND2',
    'Industrial - Warehouse': 'IND2',
    'Industrial - Heavy': 'IND1',
    'Retail': 'COM1',
    'Parking' : 'COM10'
}

convert_design_level = {
        'High-Code'    : 'HC',
        'Moderate-Code': 'MC',
        'Low-Code'     : 'LC',
        'Pre-Code'     : 'PC'
    }

def story_scale(stories, comp_type):
    if comp_type == 'NSA':
        if stories == 1:
            return 1.00
        elif stories == 2:
            return 1.22
        elif stories == 3:
            return 1.40
        elif stories == 4:
            return 1.45
        elif stories == 5:
            return 1.50
        elif stories == 6:
            return 1.90
        elif stories == 7:
            return 2.05
        elif stories == 8:
            return 2.15
        elif stories == 9:
            return 2.20
        elif (stories >= 10) and (stories < 30):
            return 2.30 + (stories-10)*0.04
        elif stories >= 30:
            return 3.10
        else:
            return 1.0

    elif comp_type in ['S', 'NSD']:
        if stories == 1:
            return 1.45
        elif stories == 2:
            return 1.90
        elif stories == 3:
            return 2.50
        elif stories == 4:
            return 2.75
        elif stories == 5:
            return 3.00
        elif stories == 6:
            return 3.50
        elif stories == 7:
            return 3.50
        elif stories == 8:
            return 3.50
        elif stories == 9:
            return 4.50
        elif (stories >= 10) and (stories < 50):
            return 4.50 + (stories-10)*0.07
        elif stories >= 50:
            return 7.30
        else:
            return 1.0

def auto_populate(DL_input_path, EDP_input_path,
                  DL_method, realization_count, coupled_EDP, event_time,
                  ground_failure, auto_script_path = None):
    """
    Short description

    Assumptions:
    - The BIM is stored under 'GeneralInformation' or 'GI' in the root of
    DL_input

    Parameters
    ----------
    DL_input_path:

    Returns
    -------
    DL_input
    DL_ap_path
    """


    # load the available DL input information
    with open(DL_input_path, 'r') as f:
        DL_input = json.load(f)

    # get the BIM data
    BIM = DL_input.get('GeneralInformation', None)
    if BIM is None:
        raise ValueError(
            "No Building Information provided for the auto-population routine."
        )

    if auto_script_path is not None: # if an external auto pop script is provided

        # load the module
        ASP = Path(auto_script_path).resolve()
        sys.path.insert(0, str(ASP.parent)+'/')
        auto_script = importlib.__import__(ASP.name[:-3], globals(), locals(), [], 0)
        auto_populate_ext = auto_script.auto_populate

        # generate the DL input data
        BIM_ap, DL_ap = auto_populate_ext(BIM = BIM)

        # add the response model information
        DL_ap.update({
            'ResponseModel': {
                'ResponseDescription': {
                    'EDP_Distribution': 'empirical',
                    'Realizations'    : realization_count
                }
            }
        })

        # add the even time information - if needed
        if (('Inhabitants' in DL_ap['LossModel'].keys()) and
            (event_time is not None)):
            DL_ap['LossModel']['Inhabitants'].update({'EventTime': event_time})

        # assemble the extended DL input
        DL_input['GeneralInformation'].update(BIM_ap)
        DL_input.update({'DamageAndLoss': DL_ap})

        # save it to the DL file with the ap suffix
        DL_ap_path = DL_input_path[:-5] + '_ap.json'

        with open(DL_ap_path, 'w') as f:
            json.dump(DL_input, f, indent=2)

        # and also return these information
        return DL_input, DL_ap_path

    else: # otherwise, use the old autopop method

        EDP_input = pd.read_csv(EDP_input_path, sep='\s+', header=0,
                                index_col=0)

        is_IM_based = DL_method[-2:] == 'IM'

        stories = BIM['NumberOfStories']
        # use only 1 story if DM is based on IM
        if DL_method == 'HAZUS MH EQ IM':
            stories = 1
        BIM.update({'NumberOfStories':stories})

        # HAZUS Earthquake
        if DL_method in ['HAZUS MH EQ', 'HAZUS MH EQ IM']:

            bt = BIM['StructureType']

            if bt == 'RV.structType':
                bt = EDP_input['structType'].values[0]

            year_built = BIM['YearBuilt']

            if bt not in ['W1', 'W2', 'S3', 'PC1', 'MH']:
                if bt not in ['URM']:
                    if stories <= 3:
                        bt += 'L'
                    elif stories <= 7:
                        bt += 'M'
                    else:
                        if bt in ['RM']:
                            bt += 'M'
                        else:
                            bt += 'H'
                else:
                    if stories <= 2:
                        bt += 'L'
                    else:
                        bt += 'M'

            if BIM['OccupancyClass'] in ap_Occupancy.keys():
                ot = ap_Occupancy[BIM['OccupancyClass']]
            else:
                ot = BIM['OccupancyClass']

            replacementCost = BIM.get('ReplacementCost', 1.0)
            replacementTime = BIM.get('ReplacementTime', 1.0)
            population = BIM.get('Population', 1.0)

            loss_dict = {
                '_method': DL_method,
                'DamageModel': {
                    'StructureType': bt
                },
                'LossModel': {
                    'DecisionVariables': {
                        'ReconstructionCost': True,
                        'ReconstructionTime': True,
                        'Injuries': True
                    },
                    'Inhabitants': {
                        'OccupancyType': ot,
                        'PeakPopulation': f'{population}'
                    },
                    'ReplacementCost': replacementCost,
                    'ReplacementTime': replacementTime
                },
                'ResponseModel': {
                    'ResponseDescription': {
                        'Realizations': realization_count,
                        "CoupledAssessment": coupled_EDP
                    }
                },
                "Dependencies": {
                    "Fragilities": "btw. Performance Groups"
                }
            }

            # add uncertainty if the EDPs are not coupled
            if not coupled_EDP:
                loss_dict['ResponseModel'].update({
                    "AdditionalUncertainty": {
                        "GroundMotion": "0.10",
                        "Modeling"    : "0.20"
                        }})

            if is_IM_based:
                loss_dict.update({
                    "ComponentDataFolder": pelicun_path+"/resources/HAZUS_MH_2.1_EQ_eqv_PGA.hdf"
                    })
            else:
                loss_dict['ResponseModel'].update({
                    'DetectionLimits': {
                        "PID": "0.20",
                        "PRD": "0.20"
                    }})
                loss_dict.update({
                    "ComponentDataFolder": pelicun_path+"/resources/HAZUS_MH_2.1_EQ_story.hdf"
                    })

            if 'W1' in bt:
                DesignL = ap_DesignLevel_W1
            else:
                DesignL = ap_DesignLevel

            for year in sorted(DesignL.keys()):
                if year_built <= year:
                    loss_dict['DamageModel'].update(
                        {'DesignLevel': DesignL[year]})
                    break

            dl = convert_design_level[loss_dict['DamageModel']['DesignLevel']]
            if 'C3' in bt:
                if dl not in ['LC', 'PC']:
                    dl = 'LC'

            # only one structural component for IM-based approach
            if is_IM_based:

                FG_S = f'S-{bt}-{dl}-{ot}'

                loss_dict.update({
                    'Components': {
                        FG_S: [
                            {'location': '1',
                             'direction': '1',
                             'median_quantity': '1.0',
                             'unit': 'ea',
                             'distribution': 'N/A'
                            }]
                    }})

            # story-based approach
            else:

                FG_S = f'S-{bt}-{dl}-{ot}'
                FG_NSD = f'NSD-{ot}'
                FG_NSA = f'NSA-{dl}-{ot}'

                loss_dict.update({
                    'Components': {
                        FG_S: [
                            {'location': 'all',
                             'direction': '1, 2',
                             #'median_quantity': '{q}'.format(q = 0.5), #/stories),
                             'median_quantity': '{q}'.format(q = story_scale(stories, 'S')/stories/2.),
                             'unit': 'ea',
                             'distribution': 'N/A'
                            }],
                        FG_NSA: [
                            {'location': 'all',
                             'direction': '1',
                             #'median_quantity': '{q}'.format(q = 1.0), #/stories),
                             'median_quantity': '{q}'.format(q = story_scale(stories, 'NSA')/stories),
                             'unit': 'ea',
                             'distribution': 'N/A'
                            }],
                        FG_NSD: [
                            {'location': 'all',
                             'direction': '1, 2',
                             #'median_quantity': '{q}'.format(q = 0.5), #/stories),
                             'median_quantity': '{q}'.format(q = story_scale(stories, 'NSD')/stories/2.),
                             'unit': 'ea',
                             'distribution': 'N/A'
                            }]
                    }})

            # if damage from ground failure is included
            if ground_failure:

                foundation_type = 'S'

                FG_GF_H = f'GF-H_{foundation_type}-{bt}'
                FG_GF_V = f'GF-V_{foundation_type}-{bt}'

                loss_dict['Components'].update({
                    FG_GF_H: [
                        {'location': '1',
                         'direction': '1',
                         'median_quantity': '1.0',
                         'unit': 'ea',
                         'distribution': 'N/A'
                        }],
                    FG_GF_V: [
                        {'location': '1',
                         'direction': '3',
                         'median_quantity': '1.0',
                         'unit': 'ea',
                         'distribution': 'N/A'
                        }]
                })

                # define logic that connects ground failure with building damage
                loss_dict.update({
                    'DamageLogic': [
                        {'type': 'propagate',
                         'source_FG': FG_GF_H,
                         'target_FG': FG_S,
                         'DS_links': {
                             '1_1': '3_1',
                             '2_1': '4_1',
                             '2_2': '4_2'
                         }
                        },
                        {'type': 'propagate',
                         'source_FG': FG_GF_V,
                         'target_FG': FG_S,
                         'DS_links': {
                             '1_1': '3_1',
                             '2_1': '4_1',
                             '2_2': '4_2'
                         }
                        }
                    ]
                })

        # HAZUS Hurricane
        elif DL_method == 'HAZUS MH HU':

            #TODO: use the HU NJ autopop script by default
            pass

        elif DL_method == 'FEMA P58':
            if BIM.get('AssetType',None) == 'Water_Pipe':

                material = BIM['Material']

                if material in ['Asbestos cement', 'Cast iron']:
                    # brittle material
                    config = 'P0001a'
                else:
                    # ductile material
                    config = 'P0001b'

                segment_count = BIM['SegmentCount']
                segment_length = BIM['Segments'][0]['length']
                cg_count = int(segment_length / (100 * ft))
                quantities = '1'
                for s in range(1, cg_count):
                    quantities += ', 1'

                loss_dict = {
                    "_method"            : "FEMA P58",
                    "ResponseModel"      : {
                        "ResponseDescription": {
                            "EDP_Distribution" : "empirical",
                            "Realizations"     : "1000",   # need to fix this later
                            "CoupledAssessment": True
                        }
                    },
                    "DamageModel"        : {
                        "CollapseProbability": {
                            "Value": "0.0",
                        },
                    },
                    "LossModel"          : {
                        "ReplacementCost"  : "1",
                        "ReplacementTime"  : "180",
                        "DecisionVariables": {
                            "Injuries"          : False,
                            "ReconstructionCost": True,
                            "ReconstructionTime": True,
                            "RedTag"            : False
                        },
                    },
                    "Dependencies"       : {
                        "CostAndTime"        : True,
                        "Fragilities"        : "btw. Damage States",
                        "Quantities"         : "Independent",
                        "ReconstructionCosts": "Independent",
                        "ReconstructionTimes": "Independent",
                    },
                    "ComponentDataFolder": "c:/Adam/Dropbox/Kutatas/2019 SC Testbeds/Memphis/",
                    "Components"         : {
                        config: [
                            {
                                "location"       : "all",
                                "direction"      : "1",
                                "median_quantity": quantities,
                                "unit"           : "ea",
                                "distribution"   : "N/A",
                            }
                        ],
                    }
                }

        if (('Inhabitants' in loss_dict['LossModel'].keys()) and
            (event_time is not None)):
                loss_dict['LossModel']['Inhabitants'].update({'EventTime': event_time})

        DL_input.update({'DamageAndLoss':loss_dict})

        DL_ap_path = DL_input_path[:-5]+'_ap.json'

        with open(DL_ap_path, 'w') as f:
            json.dump(DL_input, f, indent = 2)

        return DL_input, DL_ap_path
